﻿
namespace ImmutablePeople {
	abstract class Person {
		public string Name { get; }
		public string Password { get; }

		public Person WithName(string newName) => ClonePersonWith(newName, Password);
		public Person WithPassword(string newPassword) => ClonePersonWith(Name, newPassword);

		protected Person(string name, string password) {
			Name = name;
			Password = password;
		}

		protected abstract Person ClonePersonWith(string name, string password);
	}

	abstract class ConcretePersonBase<TConcretePerson> : Person where TConcretePerson : ConcretePersonBase<TConcretePerson>, new() {
		protected ConcretePersonBase(string name, string password) : base(name, password) { }

		protected abstract override TConcretePerson ClonePersonWith(string name, string password);

		public static TConcretePerson Default { get; } = new TConcretePerson();

		public new TConcretePerson WithName(string newName) => (TConcretePerson) base.WithName(newName);
		public new TConcretePerson WithPassword(string newPassword) => (TConcretePerson) base.WithPassword(newPassword);
	}

	class Student : ConcretePersonBase<Student> {
		public DateTime DateEnrolled { get; }

		public Student WithDateEnrolled(DateTime newDateEnrolled) => new Student(Name, Password, newDateEnrolled);

		public Student(string name, string password, DateTime dateEnrolled) : base(name, password) =>
			DateEnrolled = dateEnrolled.Date;

		public Student() : this("", "", default) { }

		protected override Student ClonePersonWith(string name, string password) =>
			new Student(name, password, this.DateEnrolled);
	}

	class Teacher : ConcretePersonBase<Teacher> {
		public int CoursesHeld { get; }

		public Teacher WithCoursesHeld(int newCoursesHeld) => new Teacher(Name, Password, newCoursesHeld);

		public Teacher(string name, string password, int coursesHeld) : base(name, password)
			=> CoursesHeld = coursesHeld;

		public Teacher() : this("", "", 0) { }

		protected override Teacher ClonePersonWith(string name, string password) =>
			new Teacher(name, password, this.CoursesHeld);
	}

	class Program {
		static void Main(string[] args) {
			var studentA = Student.Default
							.WithName("Pavel Rozek").WithPassword("lavice")
							.WithDateEnrolled(new DateTime(2015, 9, 21));

			var studentB = Student.Default
							.WithName("Marie Pichova").WithPassword("zidle")
							.WithDateEnrolled(new DateTime(2018, 9, 25));

			var teacher = Teacher.Default
							.WithName("Pavel Jezek").WithPassword("stul")
							.WithCoursesHeld(5);

			List<Person> people = new List<Person> { studentA, studentB, teacher };
			Console.WriteLine("+++ People:");
			people.PrintAll();
			Console.WriteLine();

			var updatedPeople = people.WithPasswordResetByFirstName(firstName: "Pavel", newPassword: "pohovka");
			Console.WriteLine("+++ Updated people:");
			updatedPeople.PrintAll();
			Console.WriteLine();

			Console.WriteLine("+++ People:");
			people.PrintAll();
			Console.WriteLine();

			Console.WriteLine("+++ Just students:");
			var justStudents = new List<Student> { studentA, studentB };
			justStudents.PrintAll();
			Console.WriteLine();

			var updatedStudents = justStudents.WithPasswordResetByFirstName(firstName: "Marie", newPassword: "letadlo");
			Console.WriteLine("+++ Updated students:");
			updatedStudents.PrintAll();
			Console.WriteLine();

			// Advanced: not part of original skeleton:
			var teacher2 = Teacher.Default
							.WithPassword("skrin")
							.WithCoursesHeld(5)
							.WithName("Petr Hnetynka");

			List<Person> teachers = [teacher, teacher2];
			Console.WriteLine("+++ Teachers:");
			teachers.PrintAll();
			Console.WriteLine();

			// More advanced:
			var alternativeUpdatedPeople = justStudents.AlternativeWithPasswordResetByFirstName(firstName: "Pavel", newPassword: "gauc");
			Console.WriteLine("+++ Alternative updated students:");
			alternativeUpdatedPeople.PrintAll();
			Console.WriteLine();
		}
	}

	static class ListPersonExtensions {
		public static void PrintAll(this IEnumerable<Person> people) {
			foreach (var person in people) {
				Console.WriteLine($"{person.GetType().Name} {person.Name} has password \"{person.Password}\"");
			}
		}

		public static List<Person> WithPasswordResetByFirstName(this IEnumerable<Person> people, string firstName, string newPassword) {
			var firstNameWithSpace = firstName + ' ';
			var result = new List<Person>();

			foreach (var person in people) {
				if (person.Name.StartsWith(firstNameWithSpace)) {
					result.Add(person.WithPassword(newPassword));
				} else {
					result.Add(person);
				}
			}

			return result;
		}

		public static List<TPerson> AlternativeWithPasswordResetByFirstName<TPerson>(this IEnumerable<TPerson> people, string firstName, string newPassword)
			where TPerson : Person
			{

			var firstNameWithSpace = firstName + ' ';
			var result = new List<TPerson>();

			foreach (var person in people) {
				if (person.Name.StartsWith(firstNameWithSpace)) {
					result.Add((TPerson) person.WithPassword(newPassword));
				} else {
					result.Add(person);
				}
			}

			return result;
		}
	}
}